@@ -21,31 +21,42 @@ from kazoo import client
 from kazoo import exceptions
 from kazoo.protocol import paths
 import six
-from zake import fake_client
 
 from tooz import coordination
 from tooz import locking
+from tooz import utils
 
 
 class ZooKeeperLock(locking.Lock):
-    def __init__(self, lock):
+    def __init__(self, name, lock):
+        super(ZooKeeperLock, self).__init__(name)
         self._lock = lock
 
-    def acquire(self, blocking=True, timeout=None):
-        return self._lock.acquire(blocking=blocking,
-                                  timeout=timeout)
+    def acquire(self, blocking=True):
+        return self._lock.acquire(blocking=bool(blocking),
+                                  timeout=blocking)
 
     def release(self):
         return self._lock.release()
 
 
 class BaseZooKeeperDriver(coordination.CoordinationDriver):
+    """Initialize the zookeeper driver.
+
+    :param timeout: connection timeout to wait when first connecting to the
+                    zookeeper server
+    """
 
     _TOOZ_NAMESPACE = b"tooz"
 
-    def start(self, timeout=10):
+    def __init__(self, member_id, parsed_url, options):
+        super(BaseZooKeeperDriver, self).__init__()
+        self._member_id = member_id
+        self.timeout = int(options.get('timeout', ['10'])[-1])
+
+    def _start(self):
         try:
-            self._coord.start(timeout=timeout)
+            self._coord.start(timeout=self.timeout)
         except self._coord.handler.timeout_exception as e:
             raise coordination.ToozConnectionError("operation error: %s" % (e))
 
@@ -55,39 +66,46 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             raise coordination.ToozError("operation error: %s" % (e))
 
         self._group_members = collections.defaultdict(set)
-        self._watchers = six.moves.queue.Queue()
+        self._watchers = collections.deque()
         self._leader_locks = {}
 
-    def stop(self):
+    def _stop(self):
         self._coord.stop()
 
     @staticmethod
-    def _create_group_handler(async_result, timeout, group_id):
+    def _create_group_handler(async_result, timeout,
+                              timeout_exception, group_id):
         try:
             async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NodeExistsError:
             raise coordination.GroupAlreadyExist(group_id)
         except exceptions.NoNodeError:
             raise coordination.ToozError("tooz namespace has not been created")
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
 
     def create_group(self, group_id):
         group_path = self._path_group(group_id)
         async_result = self._coord.create_async(group_path)
         return ZooAsyncResult(async_result, self._create_group_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id)
 
     @staticmethod
-    def _join_group_handler(async_result, timeout, group_id, member_id):
+    def _join_group_handler(async_result, timeout,
+                            timeout_exception, group_id, member_id):
         try:
             async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NodeExistsError:
             raise coordination.MemberAlreadyExist(group_id, member_id)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
 
     def join_group(self, group_id, capabilities=b""):
         member_path = self._path_member(group_id, self._member_id)
@@ -95,31 +113,39 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                                                 value=capabilities,
                                                 ephemeral=True)
         return ZooAsyncResult(async_result, self._join_group_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
     @staticmethod
-    def _leave_group_handler(async_result, timeout, group_id, member_id):
+    def _leave_group_handler(async_result, timeout,
+                             timeout_exception, group_id, member_id):
         try:
             async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
 
     def leave_group(self, group_id):
         member_path = self._path_member(group_id, self._member_id)
         async_result = self._coord.delete_async(member_path)
         return ZooAsyncResult(async_result, self._leave_group_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
     @staticmethod
-    def _get_members_handler(async_result, timeout, group_id):
+    def _get_members_handler(async_result, timeout,
+                             timeout_exception, group_id):
         try:
             members_ids = async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
         else:
             return set(m.encode('ascii') for m in members_ids)
 
@@ -127,33 +153,40 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         group_path = self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
         async_result = self._coord.get_children_async(group_path)
         return ZooAsyncResult(async_result, self._get_members_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id)
 
     @staticmethod
-    def _update_capabilities_handler(async_result, timeout, group_id,
-                                     member_id):
+    def _update_capabilities_handler(async_result, timeout,
+                                     timeout_exception, group_id, member_id):
         try:
             async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
 
     def update_capabilities(self, group_id, capabilities):
         member_path = self._path_member(group_id, self._member_id)
         async_result = self._coord.set_async(member_path, capabilities)
         return ZooAsyncResult(async_result, self._update_capabilities_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
     @staticmethod
-    def _get_member_capabilities_handler(async_result, timeout, group_id,
+    def _get_member_capabilities_handler(async_result, timeout,
+                                         timeout_exception, group_id,
                                          member_id):
         try:
             capabilities = async_result.get(block=True, timeout=timeout)[0]
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
         else:
             return capabilities
 
@@ -162,23 +195,27 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         async_result = self._coord.get_async(member_path)
         return ZooAsyncResult(async_result,
                               self._get_member_capabilities_handler,
+                              timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
     @staticmethod
-    def _get_groups_handler(async_result, timeout):
+    def _get_groups_handler(async_result, timeout, timeout_exception):
         try:
             group_ids = async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
         except exceptions.NoNodeError:
             raise coordination.ToozError("tooz namespace has not been created")
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(str(e))
+            raise coordination.ToozError(utils.exception_message(e))
         else:
             return set(g.encode('ascii') for g in group_ids)
 
     def get_groups(self):
         tooz_namespace = self.paths_join("/", self._TOOZ_NAMESPACE)
         async_result = self._coord.get_children_async(tooz_namespace)
-        return ZooAsyncResult(async_result, self._get_groups_handler)
+        return ZooAsyncResult(async_result, self._get_groups_handler,
+                              timeout_exception=self._timeout_exception)
 
     def _path_group(self, group_id):
         return self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
@@ -201,20 +238,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 class KazooDriver(BaseZooKeeperDriver):
     """The driver using the Kazoo client against real ZooKeeper servers."""
 
-    def __init__(self, member_id, hosts="127.0.0.1:2181", handler=None,
-                 **kwargs):
-        """:param hosts: the list of zookeeper servers in the
-        form "ip:port2, ip2:port2".
-
-        :param handler: a kazoo async handler to use if provided, if not
-        provided the default that kazoo uses internally will be used instead.
-        """
-
-        if not all((hosts, member_id)):
-            raise KeyError("hosts=%r, member_id=%r" % hosts, member_id)
+    def __init__(self, member_id, parsed_url, options):
+        super(KazooDriver, self).__init__(member_id, parsed_url, options)
+        self._coord = self._make_client(parsed_url, options)
         self._member_id = member_id
-        self._coord = client.KazooClient(hosts=hosts, handler=handler)
-        super(KazooDriver, self).__init__()
+        self._timeout_exception = self._coord.handler.timeout_exception
+
+    @classmethod
+    def _make_client(cls, parsed_url, options):
+        return client.KazooClient(hosts=parsed_url.netloc)
 
     def _watch_group(self, group_id):
         get_members_req = self.get_members(group_id)
@@ -230,7 +262,7 @@ class KazooDriver(BaseZooKeeperDriver):
                 # Copy function in case it's removed later from the
                 # hook list
                 hooks = copy.copy(self._hooks_join_group[group_id])
-                self._watchers.put(
+                self._watchers.append(
                     lambda: hooks.run(
                         coordination.MemberJoinedGroup(
                             group_id,
@@ -240,7 +272,7 @@ class KazooDriver(BaseZooKeeperDriver):
                 # Copy function in case it's removed later from the
                 # hook list
                 hooks = copy.copy(self._hooks_leave_group[group_id])
-                self._watchers.put(
+                self._watchers.append(
                     lambda: hooks.run(
                         coordination.MemberLeftGroup(
                             group_id,
@@ -332,17 +364,15 @@ class KazooDriver(BaseZooKeeperDriver):
 
     def get_lock(self, name):
         return ZooKeeperLock(
+            name,
             self._coord.Lock(
                 self.paths_join(b"/", self._TOOZ_NAMESPACE, b"locks", name),
                 self._member_id.decode('ascii')))
 
     def run_watchers(self):
         ret = []
-        while True:
-            try:
-                cb = self._watchers.get(block=False)
-            except six.moves.queue.Empty:
-                break
+        while self._watchers:
+            cb = self._watchers.popleft()
             ret.extend(cb())
 
         for group_id in six.iterkeys(self._hooks_elected_leader):
@@ -356,58 +386,14 @@ class KazooDriver(BaseZooKeeperDriver):
         return ret
 
 
-class ZakeDriver(BaseZooKeeperDriver):
-    """The driver using the Zake client which mimic a fake Kazoo client
-    without the need of real ZooKeeper servers.
-    """
-
-    def __init__(self, member_id, storage=None, **kwargs):
-        """:param storage: a fake storage object."""
-
-        if not all((storage, member_id)):
-            raise KeyError("storage=%r, member_id=%r" % storage, member_id)
-        self._member_id = member_id
-        self._coord = fake_client.FakeClient(storage=storage)
-        super(ZakeDriver, self).__init__()
-
-    @staticmethod
-    def watch_join_group(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def unwatch_join_group(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def watch_leave_group(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def unwatch_leave_group(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def watch_elected_as_leader(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def unwatch_elected_as_leader(group_id, callback):
-        raise NotImplementedError
-
-    @staticmethod
-    def run_watchers():
-        raise NotImplementedError
-
-
 class ZooAsyncResult(coordination.CoordAsyncResult):
+    def __init__(self, kazoo_async_result, handler, **kwargs):
+        self._kazoo_async_result = kazoo_async_result
+        self._handler = handler
+        self._kwargs = kwargs
 
-    def __init__(self, kazooAsyncResult, handler, **kwargs):
-        self.kazooAsyncResult = kazooAsyncResult
-        self.handler = handler
-        self.kwargs = kwargs
-
-    def get(self, timeout=15):
-        return self.handler(self.kazooAsyncResult, timeout, **self.kwargs)
+    def get(self, timeout=10):
+        return self._handler(self._kazoo_async_result, timeout, **self._kwargs)
 
     def done(self):
-        return self.kazooAsyncResult.ready()
+        return self._kazoo_async_result.ready()
