@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-#    Copyright (C) 2013 eNovance Inc. All Rights Reserved.
+#    Copyright © 2013-2014 eNovance Inc. All Rights Reserved.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -13,48 +13,43 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
-
-import threading
 import time
 import uuid
 
 import testscenarios
+from testtools import matchers
 from testtools import testcase
 
 import tooz.coordination
 from tooz import tests
-from zake import fake_storage
-
-# Real ZooKeeper server scenario
-zookeeper_tests = ('zookeeper_tests', {'backend': 'kazoo',
-                                       'kwargs': {'hosts': '127.0.0.1:2181'}})
-
-# Fake Kazoo client scenario
-fake_storage = fake_storage.FakeStorage(threading.RLock())
-fake_zookeeper_tests = ('fake_zookeeper_tests', {'backend': 'zake',
-                                                 'kwargs': {'storage':
-                                                            fake_storage}})
 
 
 class TestAPI(testscenarios.TestWithScenarios,
               tests.TestCaseSkipNotImplemented):
 
     scenarios = [
-        zookeeper_tests,
-        fake_zookeeper_tests,
-        ('memcached', {'backend': 'memcached',
-                       'kwargs': {'membership_timeout': 5}}),
+        ('zookeeper', {'url': 'kazoo://127.0.0.1:2181?timeout=5'}),
+        ('zake', {'url': 'zake://?timeout=5'}),
+        ('memcached', {'url': 'memcached://?timeout=5'}),
+        ('ipc', {'url': 'ipc://'}),
+        ('redis', {'url': 'redis://localhost:6379?timeout=5'}),
     ]
 
+    def assertRaisesAny(self, exc_classes, callable_obj, *args, **kwargs):
+        checkers = [matchers.MatchesException(exc_class)
+                    for exc_class in exc_classes]
+        matcher = matchers.Raises(matchers.MatchesAny(*checkers))
+        callable_obj = testcase.Nullary(callable_obj, *args, **kwargs)
+        self.assertThat(callable_obj, matcher)
+
     def setUp(self):
         super(TestAPI, self).setUp()
         self.group_id = self._get_random_uuid()
         self.member_id = self._get_random_uuid()
-        self._coord = tooz.coordination.get_coordinator(self.backend,
-                                                        self.member_id,
-                                                        **self.kwargs)
+        self._coord = tooz.coordination.get_coordinator(self.url,
+                                                        self.member_id)
         try:
-            self._coord.start(timeout=5)
+            self._coord.start()
         except tooz.coordination.ToozConnectionError as e:
             raise testcase.TestSkipped(str(e))
 
@@ -62,6 +57,12 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.stop()
         super(TestAPI, self).tearDown()
 
+    def test_stop_first(self):
+        c = tooz.coordination.get_coordinator(self.url,
+                                              self.member_id)
+        self.assertRaises(tooz.coordination.ToozError,
+                          c.stop)
+
     def test_create_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
@@ -96,9 +97,8 @@ class TestAPI(testscenarios.TestWithScenarios,
     def test_join_group_with_member_id_already_exists(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
-        client = tooz.coordination.get_coordinator(self.backend,
-                                                   self.member_id,
-                                                   **self.kwargs)
+        client = tooz.coordination.get_coordinator(self.url,
+                                                   self.member_id)
         client.start()
         join_group = client.join_group(self.group_id)
         self.assertRaises(tooz.coordination.MemberAlreadyExist,
@@ -122,14 +122,10 @@ class TestAPI(testscenarios.TestWithScenarios,
         all_group_ids = self._coord.get_groups().get()
         self.assertTrue(self.group_id not in all_group_ids)
         leave_group = self._coord.leave_group(self.group_id)
-        try:
-            leave_group.get()
         # Drivers raise one of those depending on their capability
-        except (tooz.coordination.MemberNotJoined,
-                tooz.coordination.GroupNotCreated):
-            pass
-        else:
-            self.fail("Exception not raised")
+        self.assertRaisesAny([tooz.coordination.MemberNotJoined,
+                              tooz.coordination.GroupNotCreated],
+                             leave_group.get)
 
     def test_leave_group_not_joined_by_member(self):
         self._coord.create_group(self.group_id).get()
@@ -142,9 +138,8 @@ class TestAPI(testscenarios.TestWithScenarios,
     def test_get_members(self):
         group_id_test2 = self._get_random_uuid()
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
 
         self._coord.create_group(group_id_test2).get()
@@ -165,14 +160,10 @@ class TestAPI(testscenarios.TestWithScenarios,
     def test_get_member_capabilities_nonexistent_group(self):
         capa = self._coord.get_member_capabilities(self.group_id,
                                                    self.member_id)
-        try:
-            capa.get()
         # Drivers raise one of those depending on their capability
-        except (tooz.coordination.MemberNotJoined,
-                tooz.coordination.GroupNotCreated):
-            pass
-        else:
-            self.fail("Exception not raised")
+        self.assertRaisesAny([tooz.coordination.MemberNotJoined,
+                              tooz.coordination.GroupNotCreated],
+                             capa.get)
 
     def test_get_member_capabilities_nonjoined_member(self):
         self._coord.create_group(self.group_id).get()
@@ -198,25 +189,18 @@ class TestAPI(testscenarios.TestWithScenarios,
     def test_update_capabilities_with_group_id_nonexistent(self):
         update_cap = self._coord.update_capabilities(self.group_id,
                                                      b'test_capabilities')
-        try:
-            update_cap.get()
         # Drivers raise one of those depending on their capability
-        except (tooz.coordination.MemberNotJoined,
-                tooz.coordination.GroupNotCreated):
-            pass
-        else:
-            self.fail("Exception not raised")
+        self.assertRaisesAny([tooz.coordination.MemberNotJoined,
+                              tooz.coordination.GroupNotCreated],
+                             update_cap.get)
 
     def test_heartbeat(self):
         self._coord.heartbeat()
 
     def test_disconnect_leave_group(self):
-        if self.backend == 'zake':
-            self.skipTest("Zake has a bug that prevent this test from working")
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -230,12 +214,11 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertTrue(member_id_test2 not in members_ids)
 
     def test_timeout(self):
-        if self.backend != 'memcached':
+        if not self.url.startswith('memcached://'):
             self.skipTest("This test only works with memcached for now")
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -256,9 +239,8 @@ class TestAPI(testscenarios.TestWithScenarios,
 
     def test_watch_group_join(self):
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
 
@@ -291,9 +273,8 @@ class TestAPI(testscenarios.TestWithScenarios,
 
     def test_watch_leave_group(self):
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
 
@@ -365,9 +346,8 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.run_watchers()
 
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
@@ -395,6 +375,9 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertEqual(client2.get_leader(self.group_id).get(),
                          member_id_test2)
 
+        # Restart the coord because tearDown stops it
+        self._coord.start()
+
     def test_get_leader(self):
         self._coord.create_group(self.group_id).get()
 
@@ -419,9 +402,8 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.run_watchers()
 
         member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id_test2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id_test2)
         client2.start()
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
@@ -466,9 +448,8 @@ class TestAPI(testscenarios.TestWithScenarios,
 
     def test_get_lock_multiple_coords(self):
         member_id2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.backend,
-                                                    member_id2,
-                                                    **self.kwargs)
+        client2 = tooz.coordination.get_coordinator(self.url,
+                                                    member_id2)
         client2.start()
 
         lock_name = self._get_random_uuid()
